import re
from pathlib import Path
import asyncio
import json
import pandas as pd
from dl2050utils.core import *

def q(db, qss):
    res = None
    for qs in qss.split(';'): res = asyncio.get_event_loop().run_until_complete(db.query(qs))
    return res

def get_model_names(path):
    models = []
    for f in Path(path).glob('*.json'):
        r = re.search('(.*)$', f.stem, re.IGNORECASE)
        if(r is None): continue
        models.append(r.group(1))
    return models

def create_model(db, model, fname=None, path='.'):
    fname = fname or f'{model}.json'
    fname = f'{path}/{fname}'
    try:
        with open(fname) as f: meta=json.load(f)
        q(db, f"insert into models (model,meta) values ('{model}','{json.dumps(meta)}')")
        return False
    except Exception as e:
        print(str(e))
        return True

def update_model(db, model, fname=None, path='.'):
    fname = fname or f'{model}.json'
    fname = f'{path}/{fname}'
    try:
        with open(fname) as f: meta=json.load(f)
        q(db, f"update models set meta='{json.dumps(meta)}' where model='{model}'")
        return False
    except Exception as e:
        print(str(e))
        return True
    
def merge_model(db, model, fname, path='.'):
    fname = f'{path}/{fname}'
    try:
        meta = get_model(db, model)
        with open(fname) as f: meta2=json.load(f)
        meta = {**meta, **meta2}
        q(db, f"update models set meta='{json.dumps(meta)}' where model='{model}'")
        return False
    except Exception as e:
        print(str(e))
        return True

def delete_model(db, model):
    return q(db, f"delete from models where model='{model}'")

def get_all_models(db):
    return [e['model'] for e in q(db,f"select model from models")]

def get_model(db, model):
    try:
        return json.loads(q(db, f"select meta from models where model='{model}'")[0]['meta'])
    except Exception as e:
        return

async def db_insert_df(db, df, tbl, renames={}):
    for _,row in df.iterrows():
        d = {}
        for k,v in row.items(): d[renames[k] if k in renames else k] = v
        await db.insert(tbl, d)

def get_tree_rows(rows, idx, parent, df, cols, level, crank=None, max_level=None):
    if max_level and level>=max_level: return 0
    c1 = cols[level]
    vals = df.groupby(c1).sum().sort_values(crank,ascending=False).index.values
    for v in vals:
        df1 = df[df[c1]==v]
        nodeidx = idx[0]
        idx[0] += 1
        n = get_tree_rows(rows, idx, nodeidx, df1, cols, level+1, crank=crank, max_level=max_level)
        nn = idx[0]-nodeidx-1
        leaf = n==0
        rows.append([nodeidx, parent, level+1, leaf, n, nn, v])
    return len(vals)

def build_tree_tbl(df, cols, vcol, crank=None, max_level=None):
    df1 = df[cols+[vcol]].groupby(cols).sum().reset_index()
    for c in cols: df1.loc[df1[c]=='',c] = 'Etc'
    tcols = ['id', 'parent', 'lv', 'leaf', 'n', 'nn', 'name']
    rows = []
    get_tree_rows(rows, [1], 0, df1, cols, 0, crank=crank, max_level=max_level)
    return pd.DataFrame(rows, columns=tcols).sort_values('id')

def get_tree_json(df, parent=0):
    rows = []
    df1 = df[df['parent']==parent]
    for i in range(len(df1)):
        d = df1.iloc[i].to_dict()
        if d['leaf']!=True: d['children']=get_tree_json(df, parent=d['id'])
        rows.append(d)
    return rows

def get_json(df, cols, key):
    res = []
    while True:
        df1 = df[df['id']==key]
        if len(df1)==0: break
        d = df1.iloc[0].to_dict()
        level = d['level']
        if level>len(cols) or level<0: break
        res.append({'col': cols[level-1], 'id': d['id'], 'name': d['name']})
        key = d['parent']
    return res



#
# dbutils.py
#

from dl2050utils.core import oget, check_str, is_numeric_str, check

# ####################################################################################################
# Utils
# ####################################################################################################

def str_to_int_or_float(s):
    """
    Convert a string to an int or float depending on the contetent.
    If not a numeric string return the string as is.
    """
    if not is_numeric_str(s): return s
    v = float(s)
    if int(v)==v: return int(v)
    return v

def is_valid_list(l):
    """Returns True if l is a list and all elements are basic elements, False otherwise"""
    if type(l)!=list: return False
    allowed_types = (int, float, str, bool, type(None))
    return True if all(isinstance(e, allowed_types) for e in l) else False

def convert_filter_val(val, max_sz=1024, none_val='NULL'):
    """
    """
    if val is None: return none_val
    if type(val) in [bool,int,float]: return val
    if is_numeric_str(val): return  str_to_int_or_float(val)
    if type(val)==str: return val[:max_sz]
    return None

# ####################################################################################################
# parse_filter
# ####################################################################################################

def parse_filter(f, max_sz=1024):
    """
    Checks if a filter is valid.
    The filter is valid if it contains col,val attributes.
    col attribute must be string.
    val attributes can be string, int, float or lists of basic elements, only for IN and NOT IN operators.
    String attributes are enforce to max lengh of max_sz.
    Returns the filter cleared of extra attributes or None if error.
    """
    # Check for valid filter structure
    if type(f)!=dict: return None
    if 'col' not in f or 'val' not in f: return None
    if type(f['col'])!=str: return None
    col,val = f['col'][:max_sz],f['val']
    op = oget(f,['op'],'=').upper()
    # Check basic filters (not lists)
    if type(val)!=list:
        if op not in ['=', '!=', '>', '>=', '<', '<=']: return None
        val = convert_filter_val(val, max_sz=max_sz)
        if val=='NULL':
            if op=='!=': val = 'NOT NULL'
            op = ' IS '
        return {'col':col, 'val':val, 'op':op}
    # List values are only allowed for IN and NOT_IN ops
    if op not in ['IN', 'NOT_IN']: return None
    # List must contain only base elements
    if not is_valid_list(val): return None
    # Convert all list value elements
    val = [convert_filter_val(e) for e in val]
    return {'col':col, 'val':val, 'op':op}

# ####################################################################################################
# parse_filters
# ####################################################################################################

def parse_filters(fs, max_sz=1024):
    """
    Filters can be either a list of dicts or a dict with key/vals representing filter conditions.
    A dict filter is converted into a list filters.
    All list filter elements are dicts with one filter condition defined by key/val and optionally the op attributes.
    The op attribut can be =, < , > , <=, >=, !=, IN, NOT_IN
    Examples:
        - Dict filter: {'id':1, 'name':'name1'}
        - List filter: [{'col':'id', 'val':1 'op':'='}, {'col':'name', 'val':'name1' 'op':'='}]
    """
    # Define valid ops
    # Filters must be either dict or list
    if type(fs)!=dict and type(fs)!=list: return []
    # Dict filters
    if type(fs) == dict: return [{"col":check_str(e, n=1024), "val":check(fs[e]), "op": "="} for e in fs]
    # List filters
    fs2 = []
    for f in fs:
        f2 = parse_filter(f, max_sz=max_sz)
        if f2 is None: continue
        fs2.append(f2)
    return fs2
